#include "latency_compensator.h"

#include <cstring>
#include <ranges>
#include <vector>
#include <memory>

#include "audio/device/audio_device_manager.h"
#include "extern.h"
#include "thread_message/thread_message_hubs.h"

latency_compensator::latency_compensator() {
}

latency_compensator::~latency_compensator() {
}

void latency_compensator::init() {
    const uint32_t channel_count = g_audio_device_manager.get_output_channel_count();
    internal_buffer.resize(channel_count);
	ui_buffers.resize(channel_count);

    uint32_t block_size = g_audio_device_manager.get_buffer_size();
    resize(block_size);
	for (auto& buffer : ui_buffers) {
		buffer.set_capacity(block_size * 2);
	}
}

void latency_compensator::set_latency(int32_t in_latency_offset) {
    latency_ = in_latency_offset;
    g_audio_thread_hub.push_message([]() {
        on_latency_offset_changed.broadcast();
    });
}

void latency_compensator::set_user_latency(int32_t in_latency_offset) {
    user_latency_ = in_latency_offset;
    g_audio_thread_hub.push_message([]() {
        on_latency_offset_changed.broadcast();
    });
}

void latency_compensator::push(audio_buffer& in_buffer) {
    auto headers = in_buffer.get_headers();
    for (uint32_t i = 0; i < internal_buffer.size(); ++i) {
        internal_buffer[i].push(headers[i], in_buffer.get_num_samples());
    }
}

// 1. 假设需要延后20个sample(latency_offset = 20), 当前frame_size为512
// 2. 那么在第一次pop的时候, 需要在前面补充20个0, 然后再后面补上512-20个sample

// 1. 如果要提前20个sample(latency_offset = -20), 那么就相当于除去这个混音轨道, 其他的混音轨道都延迟20个sample
void latency_compensator::pop(uint64_t delta_sample, circular_buffer_vector_type& out_buffer) {
    std::vector<sample_t> temp_buffer(delta_sample);

    for (auto buffer : std::views::zip(internal_buffer, ui_buffers, out_buffer)) {
        auto& [internal, ui_buffer, target_buffer] = buffer;
        internal.pop(temp_buffer.data(), delta_sample);
        target_buffer.push(temp_buffer);
		ui_buffer.push(temp_buffer);

        memset(temp_buffer.data(), 0, delta_sample * sizeof(sample_t));
    }
}

void latency_compensator::pop(uint64_t delta_smaple, audio_buffer& out_buffer) {
    for (auto buffer : std::views::zip(internal_buffer, ui_buffers, out_buffer.get_headers_vector())) {
        auto& [internal, ui_buffer, target_buffer] = buffer;
        internal.pop(target_buffer, delta_smaple);
		ui_buffer.push(target_buffer, delta_smaple);
    }
}

void latency_compensator::update_buffer(int32_t min_latency, int32_t max_latency) {
    // 根据全局最大延迟计算需要的buffer大小
    const uint32_t block_size = g_audio_device_manager.get_buffer_size();
    const auto& buffer_size = std::max(0, max_latency - get_real_latency());

    const int32_t block_num  = buffer_size / block_size + 1;
    resize(block_num * block_size);

    const auto& buffer_latency = max_latency - (get_real_latency() - min_latency); // 计算需要的buffer大小
    push_zeros(buffer_latency);
}
