#include "mixer_track.h"

#include "channel_interface.h"
#include "audio/device/audio_device_manager.h"
#include "audio/plugin_host/plugin_host.h"
#include "thread_message/thread_message_hubs.h"

mixer_track::~mixer_track() {
    delete channel_interface_;
    channel_interface_ = nullptr;
    for (auto e : effects) {
        // e->on_latency_changed.remove_all(this);
        delete_effect(this, e);
    }
}

void mixer_track::init() {
    const uint32_t channel_count = g_audio_device_manager.get_output_channel_count();
    buffer_.resize(channel_count, g_audio_device_manager.get_buffer_size());
    temp_mix_buffer_.resize(buffer_.get_num_channels(), g_audio_device_manager.get_buffer_size());

    channel_interface_ = new mixer_channel_interface(this);
    compensator_.init();
}

void mixer_track::add_effect(plugin_host* in_effect) {
    in_effect->on_latency_changed.add_raw(this, &mixer_track::on_latency_changed);

    g_audio_thread_hub.push_message([in_effect, this] {
        in_effect->owner_tracks.push_back(this);
        in_effect->channel->set_input_channel(channel_interface_->output_channel_nodes);
        in_effect->channel->set_output_channel(channel_interface_->output_channel_nodes);
        effects.push_back(in_effect);
    });
}

void mixer_track::remove_effect(plugin_host* in_effect) {
    g_audio_thread_hub.push_message([in_effect, this] {
        on_latency_changed(0);
        auto remove_effect_track = std::remove(in_effect->owner_tracks.begin(), in_effect->owner_tracks.end(), this);
        in_effect->owner_tracks.erase(remove_effect_track, in_effect->owner_tracks.end());

        auto remove_effect = std::remove(effects.begin(), effects.end(), in_effect);
        effects.erase(remove_effect, effects.end());
    });
}

void mixer_track::add_child(mixer_track* in_child, float in_send_level) {
    children.push_back({in_child, in_send_level});
}

void mixer_track::remove_child(mixer_track* in_child) {
    const auto new_end = std::remove_if(children.begin(), children.end(), [in_child](const mixer_track_link& link) {
        return link.track == in_child;
    });
    children.erase(new_end, children.end());
}

void mixer_track::process(uint32_t in_frames) {
    // 从children中获取buffer
    for (const auto& link : children) {
        link.track->compensator_.pop(in_frames, temp_mix_buffer_);
        // audio_buffer& child_buffer = link.track->buffer;
        buffer_.add(temp_mix_buffer_, link.send_level);
    }
    for (auto effect : effects)
        effect->process(in_frames);
	post_process(in_frames);
    compensator_.push(buffer_);
}

void mixer_track::post_process(uint32_t in_frames) {
    buffer_.multiple(get_volume());
}

void mixer_track::on_latency_changed(int32_t in_latency) {
    int32_t latency = 0;
    for (const auto effect: effects) {
        latency += effect->get_latency();
    }
    compensator_.set_latency(latency);
}

instrument_track::instrument_track(plugin_host* in_instrument): mixer_track(mixer_track_type::instrument), instrument_(in_instrument) {
    in_instrument->on_latency_changed.add_raw(this, &instrument_track::on_latency_changed);
}

void instrument_track::rename(const std::string& in_name) {
    instrument_->name = in_name;
}

std::string instrument_track::get_name() const {
    return instrument_->name;
}

void instrument_track::on_latency_changed(int32_t in_latency) {
    int32_t latency = 0;
    for (const auto effect: effects) {
        latency += effect->get_latency();
    }
    compensator_.set_latency(latency);
}

void delete_effect(mixer_track* track, plugin_host* host) {
    track->remove_effect(host);
    host->try_close_editor();
    g_audio_thread_hub.push_message([host, track](){
        delete host;
    });
}
